import time
import sys
import pymodbus

from edge_psu_ctl.transport_pymodbus import PyTransport
from edge_psu_ctl import lamp_control
from edge.com.fs import try_lock_pid_file, try_rm_pid_file
from edge.pipeline_helpers import load_pscuctl_json
from edge import eventlog

_PID_LOCK_PATH = '/var/run/edge/led-ctl.pid'
_RETRY_INTERVAL = 0.25  # seconds
_RETRY_MAX_COUNT = 3


class _AppError(RuntimeError):
    pass


class _Args:
    LED_RADAR = 'radar'
    LED_CAMERA = 'camera'
    LED_ETHERNET = 'ethernet'

    MODE_CONTINUOUS = 'cont'
    MODE_FLASH_FAST = 'flash-fast'
    MODE_FLASH_SLOW = 'flash-slow'

    COLOR_NONE = 'off'
    COLOR_RED = 'red'
    COLOR_GREEN = 'green'
    COLOR_YELLOW = 'yellow'

    def __init__(self, led, led_color, led_mode):
        """
        :type led: int
        :type led_mode: int
        :type led_color: int
        """
        self.led = led
        self.led_mode = led_mode
        self.led_color = led_color

    @staticmethod
    def deserialize(raw_args):
        """
        :type raw_args: list[str]
        :rtype: _Args
        :raises _AppError:
        """
        module_to_led_map = {
            'ethernet': lamp_control.LAMP_1RG,
            'camera': lamp_control.LAMP_2RG,
            'radar': lamp_control.LAMP_3RG
        }

        color_to_led_map = {
            'green': lamp_control.COLOR_GREEN,
            'yellow': lamp_control.COLOR_YELLOW,
            'red': lamp_control.COLOR_RED,
            'off': lamp_control.COLOR_NONE
        }

        mode_to_led_map = {
            _Args.MODE_CONTINUOUS: lamp_control.MODE_CONTINUOUS,
            _Args.MODE_FLASH_FAST: lamp_control.MODE_SWITCHING_2HZ,
            _Args.MODE_FLASH_SLOW: lamp_control.MODE_SWITCHING_1HZ
        }

        if len(raw_args) < 4:
            eventlog.log_d('fail: invalid number of arguments')
            raise _AppError

        # define led
        raw_led = str(raw_args[1]).strip()
        if raw_led in (_Args.LED_ETHERNET, _Args.LED_CAMERA, _Args.LED_RADAR):
            led = module_to_led_map[raw_led]
        else:
            eventlog.log_d('fail: param <1> has invalid value <%s>' % raw_led)
            raise _AppError

        # define led-color
        raw_color = str(raw_args[2]).strip()
        if raw_color in (_Args.COLOR_GREEN, _Args.COLOR_RED, _Args.COLOR_YELLOW, _Args.COLOR_NONE):
            led_color = color_to_led_map[raw_color]
        else:
            eventlog.log_d('fail: param <2> has invalid value <%s>' % raw_color)
            raise _AppError

        # define led-mode
        raw_mode = str(raw_args[3]).strip()
        if raw_mode in (_Args.MODE_FLASH_SLOW, _Args.MODE_FLASH_FAST, _Args.MODE_CONTINUOUS):
            led_mode = mode_to_led_map[raw_mode]
        else:
            eventlog.log_d('fail: param <3> has invalid value <%s>' % raw_mode)
            raise _AppError

        return _Args(led, led_color, led_mode)


def _get_com_endpoint():
    """
    :rtype: str | None
    """
    psuctl_conf = load_pscuctl_json()
    if not psuctl_conf:
        eventlog.log_d('fail: psu-ctl.conf not found')
        return

    ro_jnode = psuctl_conf.get('ro')
    if not ro_jnode:
        eventlog.log_d('fail: param psu-ctl.conf#/ro has invalid value')
        return

    com_port = ro_jnode.get('com-port')
    if not com_port:
        eventlog.log_d('fail: param psu-ctl.conf#/ro/com-port has invalid value')
        return

    return com_port


def _main(raw_args):
    """
    :type raw_args: list[str]
    """
    com_port = _get_com_endpoint()
    if not com_port:
        eventlog.log_d('fail: psu-ctl hardware not found')
        return

    args = _Args.deserialize(raw_args)

    retry_count = 1
    while retry_count <= _RETRY_MAX_COUNT:
        if not try_lock_pid_file(_PID_LOCK_PATH):
            eventlog.log_d('fail: another instance is running, attempt=%s' % retry_count)
            time.sleep(_RETRY_INTERVAL)
            retry_count += 1
            continue

        # noinspection PyBroadException
        try:
            transport = PyTransport(com_port)
            control = lamp_control.LampControl(transport, args.led)
            control.set_state(args.led_mode, args.led_color)
            return

        except pymodbus.exceptions.ModbusIOException:
            # workaround for:
            # pymodbus.exceptions.ModbusIOException:
            # Modbus Error: [Input/Output] Server responded with bad response
            return

        except pymodbus.exceptions.ConnectionException as x:
            eventlog.log_d('fail: %s' % x)
            retry_count += 1
            continue

        except Exception:
            eventlog.log_x()
            retry_count += 1
            continue

        finally:
            try_rm_pid_file(_PID_LOCK_PATH)

    eventlog.log_d('fail: retry count exceeded')


if __name__ == '__main__':
    eventlog.open_log('led-ctl', 'debug')
    # noinspection PyBroadException
    try:
        _main(sys.argv)
    except _AppError:
        pass
    except Exception:
        eventlog.log_x()
    finally:
        eventlog.close_log()
