#!/usr/bin/python

from edge import eventlog
from edge.admin import reply_helpers as reply
from edge.admin.cgi import run_cgi, CgiModule, CgiPathInfo
from edge.admin.cgi import CgiBadRequestError, CgiNotFoundError
from edge.com import fs

from socket import gethostname
import apt

_ABOUT_FILEPATH = '/etc/edge/setupstore/about.json'


class _About:
    def __init__(self):
        pass

    @staticmethod
    def get():
        """
        :rtype: dict
        """
        def _get_sw_version(package_names):
            """
            :type package_names: list[str]
            :rtype: list[dict]
            """
            answer = list()
            with apt.cache.Cache() as apt_cache:
                for name in package_names:
                    package = apt_cache[name]
                    version = package.installed
                    if version:
                        answer.append({
                            'package': ['VIT', name, version.version]
                        })

            return answer

        def _get_hw_version():
            """
            :rtype: dict | None
            """
            about_jnode = fs.load_json(_ABOUT_FILEPATH) or dict()
            return about_jnode.get('hw')

        device_jnode = {
            'hostname': gethostname()
        }

        hw_version = _get_hw_version()
        if hw_version:
            device_jnode['hw'] = hw_version

        sw_version = _get_sw_version(['edge'])
        if sw_version:
            device_jnode['sw'] = sw_version

        return {
            'device': device_jnode
        }


class _Module(CgiModule):
    """
    GET  /api/about  /
    """
    def __init__(self, handler):
        """
        :type handler: () -> Reply
        """
        CgiModule.__init__(self, handler)

    @classmethod
    def unpack(cls, request, about):
        """
        :type request: CgiRequest
        :type about: _About
        :rtype: CgiModule

        :raise CgiNotFoundError:
        :raise CgiBadRequestError:
        """
        method_to_handlers_map = {
            'GET': [
                cls._get
            ]
        }

        handlers = method_to_handlers_map.get(request.method, [])
        for handler in handlers:
            try:
                return handler(request, about)

            except CgiBadRequestError:
                raise

            except CgiNotFoundError:
                pass

        eventlog.log_d('fail: http404')
        raise CgiNotFoundError

    @staticmethod
    def _get(request, about):
        """
        :type request: CgiRequest
        :type about: _About
        :rtype: CgiModule
        """
        def check_path_info(path_info):
            """
            :type path_info: CgiPathInfo

            :raise CgiNotFoundError:
            """
            if len(path_info) != 0:
                eventlog.log_i('fail: http404')
                raise CgiNotFoundError

        def check_args():
            if request.query_string:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

            if request.body:
                eventlog.log_i('fail: http400')
                raise CgiBadRequestError

        check_path_info(request.path_info)
        check_args()

        def handler():
            """
            :rtype: Reply
            """
            reply_jnode = about.get()
            return reply.Reply(body=reply_jnode)

        return _Module(handler)


def _main():
    about = _About()

    run_cgi(
        lambda arg: CgiPathInfo.unpack(arg),
        lambda arg: _Module.unpack(arg, about))


if __name__ == '__main__':
    eventlog.open_log('cgi-about', 'debug')
    try:
        _main()
    finally:
        eventlog.close_log()
