#!/usr/bin/python

import sys
import fcntl
import ujson
import jsonschema
import os
import datetime
import pytz

from edge import eventlog
from edge import path_helpers as ph
from edge import pipeline_helpers as oh
from edge.com import fs
from edge.admin import reply_helpers as rh
from edge.admin import ui_types
from pipeline import _edge_blobstore as bs


UI_EVENTSTORE_BASEDIR = '/etc/edge/setupstore/admin/actions/eventstore'

UI_EVENTSTORE_CHANNEL_DEFJNODE_FILEPATH = '/opt/edge/share/setupstore/ui/eventstore-ch-jnode.json'
UI_EVENTSTORE_CHANNEL_DEFSCHEMA_FILEPATH = '/opt/edge/share/setupstore/ui/eventstore-ch-schema.json'

STAT_FILE_PATH = '/tmp/edge/zabbix/bs65535'


def _get_ch_jnode_path(ch_id):
    return fs.join_path(UI_EVENTSTORE_BASEDIR, ch_id, '0.json')


def _read_raw_body():
    """
    :rtype: str | None
    """
    # make stdin a non-blocking file
    fd = sys.stdin.fileno()
    old_flags = fcntl.fcntl(fd, fcntl.F_GETFL)
    fcntl.fcntl(fd, fcntl.F_SETFL, old_flags | os.O_NONBLOCK)

    try:
        raw_body = sys.stdin.readline()
        if raw_body:
            return raw_body

    except Exception as x:
        eventlog.log_d("fail: read raw-body (%s)" % x.message)

    finally:
        fcntl.fcntl(fd, fcntl.F_SETFL, old_flags)

    return None


def _get_args():
    try:
        http_method = os.environ['REQUEST_METHOD']
        raw_path_info = os.environ['PATH_INFO']
        if raw_path_info:
            path_info = raw_path_info.split('/')[1:]
            ch_id = raw_path_info.split('/')[1]
        else:
            path_info = []
            ch_id = None

        query_string = os.environ['QUERY_STRING']
        if not query_string:
            query_string = None

        raw_body = _read_raw_body()
        eventlog.log_d("%s|%s|%s|%s" % (http_method, raw_path_info, query_string, raw_body))
        query_args = http_method, path_info, query_string, raw_body
        return query_args

    except Exception as x:
        eventlog.log_d("fail: deserialize query-args (%s)" % x.message)
        raise


def _handle_get_eventstore(channel_id):
    """
    :type channel_id: str
    :rtype:dict
    """
    # GET [/api/v1/admin/actions/eventstore/<id>]
    evenstore_jnode_filepath = _get_ch_jnode_path(channel_id)
    if not os.path.exists(evenstore_jnode_filepath):
        return rh.reply_fail(rh.REPLY_BAD_REQUEST_400)

    ui_jnode = fs.load_json(evenstore_jnode_filepath)
    if ui_jnode is None:
        return rh.reply_fail(rh.REPLY_INTERNAL_SERVER_ERROR_500)

    stats = dict()

    time_jnode = fs.load_json(ph.UI_TIME_JNODE_FILEPATH)
    tz = None
    if time_jnode:
        tz_name = time_jnode.get('rw', dict()).get('args', dict()).get('tz')
        eventlog.log_i('tz %s' % tz_name)
        tz = pytz.timezone(tz_name)

    stat_jnode = fs.load_json(STAT_FILE_PATH)
    if stat_jnode:
        ch_stats = filter(lambda storage_jnode: storage_jnode['channel'] == channel_id, stat_jnode['storage'])
        if len(ch_stats) == 1:
            stats = {
                "time": datetime.datetime.fromtimestamp(stat_jnode['stat']['mtime'][0], tz=tz).isoformat(),
                "min-time": datetime.datetime.fromtimestamp(ch_stats[0]['min-time'][0], tz=tz).isoformat(),
                "max-time": datetime.datetime.fromtimestamp(ch_stats[0]['max-time'][0], tz=tz).isoformat(),
                "df": ch_stats[0]['df'],
                "du": ch_stats[0]['du'],
                "diru": ch_stats[0]['diru']
            }

    limits = {
        "cleanup": {
            "min-df": {
                "min": 0,
                "max": 100500
            },
            "max-du": {
                "min": 0,
                "max": 100500
            },
            "max-abs-time": {
                "min": 0,
                "max": 100500
            },
            "max-rel-time": {
                "min": 0,
                "max": 100500
            }
        }
    }
    ui_jnode['rw']['limits'] = limits
    ui_jnode['rw']['stats'] = stats
    return rh.reply_done(ui_jnode['rw'])


def _handle_get(query_args):
    """
    :type query_args: ui_types.UiQueryArgs
    :return:
    """
    if query_args.raw_body:
        eventlog.log_d('fail: bad-request')
        return rh.reply_fail(rh.REPLY_BAD_REQUEST_400)

    if len(query_args.path_info) == 0:
        # GET [/api/v1/admin/actions/]

        ui_jnode = oh.get_ui_pipeline_eventstore_jnode()
        if ui_jnode:
            return rh.reply_done({
                'eventstores': ui_jnode
            })
        else:
            return rh.reply_fail(rh.REPLY_INTERNAL_SERVER_ERROR_500)

    elif len(query_args.path_info) == 2:
        module_type, module_id = query_args.path_info
        if module_type == 'eventstore':
            return _handle_get_eventstore(module_id)
        else:
            return rh.reply_fail(rh.REPLY_BAD_REQUEST_400)


def _handle_put_eventstore(query_args):
    # PUT /api/v1/admin/actions/eventstore/<id>
    try:
        new_ui_jnode = ujson.loads(query_args.raw_body) if query_args.raw_body else None
    except ValueError as x:
        eventlog.log_d("fail: bad-request (%s)" % x.message)
        return rh.reply_fail(rh.REPLY_BAD_REQUEST_400)

    _, channel_id = query_args.path_info

    if not channel_id:
        eventlog.log_d('fail: put without channel name')
        return rh.reply_fail(rh.REPLY_BAD_REQUEST_400)

    evenstore_jnode_filepath = fs.join_path(UI_EVENTSTORE_BASEDIR, channel_id, '0.json')

    raw_ui_jnode = fs.load_json(evenstore_jnode_filepath)
    if raw_ui_jnode is None:
        return rh.reply_fail(rh.REPLY_INTERNAL_SERVER_ERROR_500)

    request_schema_jnode = fs.load_json2(UI_EVENTSTORE_CHANNEL_DEFSCHEMA_FILEPATH)
    try:
        jsonschema.validate(new_ui_jnode, request_schema_jnode)
    except (ValueError, jsonschema.ValidationError) as x:
        eventlog.log_d('fail: schema validation [%s]' % x.message)
        return rh.reply_fail(rh.REPLY_BAD_REQUEST_400)

    if new_ui_jnode['args']['name'] != raw_ui_jnode['rw']['args']['name']:
        # change store name in /etc/edge/setupstore/admin/action/eventstore.json
        raw_ui_set_jnode = oh.get_ui_pipeline_eventstore_jnode()
        if raw_ui_set_jnode is None:
            return rh.reply_fail(rh.REPLY_INTERNAL_SERVER_ERROR_500)

        store_jnode_set = filter(lambda item_jnode: item_jnode['id'] == channel_id, raw_ui_set_jnode)
        if len(store_jnode_set) != 1:
            eventlog.log_d("fail: internal-server-error")
            return rh.reply_fail(rh.REPLY_INTERNAL_SERVER_ERROR_500)

        store_jnode_set[0]['name'] = new_ui_jnode['args']['name']

        oh.set_ui_pipeline_eventstore_jnode(raw_ui_set_jnode)

    raw_ui_jnode['rw']['args'] = new_ui_jnode['args']

    bs.reconfigure_blobstore(raw_ui_jnode)

    with open(evenstore_jnode_filepath, 'w') as json_file:
        ujson.dump(raw_ui_jnode, json_file, indent=2)

    stats = {}
    stat_jnode = fs.load_json(STAT_FILE_PATH)
    if stat_jnode:
        ch_stats = filter(lambda storage_jnode: storage_jnode['channel'] == channel_id, stat_jnode['storage'])
        if len(ch_stats) == 1:
            stats = {
                "time": datetime.datetime.fromtimestamp(stat_jnode['stat']['mtime'][0]).isoformat(),
                "min-time": datetime.datetime.fromtimestamp(ch_stats[0]['min-time'][0]).isoformat(),
                "max-time": datetime.datetime.fromtimestamp(ch_stats[0]['max-time'][0]).isoformat(),
                "df": ch_stats[0]['df'],
                "du": ch_stats[0]['du'],
                "diru": ch_stats[0]['diru']
            }

    limits = {
        "cleanup": {
            "min-df": {
                "min": 0,
                "max": 100500
            },
            "max-du": {
                "min": 0,
                "max": 100500
            },
            "max-abs-time": {
                "min": 0,
                "max": 100500
            },
            "max-rel-time": {
                "min": 0,
                "max": 100500
            }
        }
    }
    raw_ui_jnode['rw']['limits'] = limits
    raw_ui_jnode['rw']['stats'] = stats
    return rh.reply_done(raw_ui_jnode['rw'])


def _get_new_channel_id():
    """
    :rtype: str | None
    """
    ui_jnode = oh.get_ui_pipeline_eventstore_jnode()
    if ui_jnode is None:
        eventlog.log_d('fail: load ui-jnode')
        return None

    used_ch_id_set = set(map(lambda eventstore_jnode: int(eventstore_jnode['id']), ui_jnode))

    ch_id = None
    for ch_id_candidate in xrange(0, 32):
        if ch_id_candidate not in used_ch_id_set:
            ch_id = str(ch_id_candidate)
            eventlog.log_d("new-channel-id: %s" % ch_id)
            break
    return ch_id


def _handle_post_eventstore(query_args):
    # POST /api/v1/admin/actions/eventstore
    try:
        new_jnode = ujson.loads(query_args.raw_body) if query_args.raw_body else None
    except ValueError as x:
        eventlog.log_d("fail: bad-request (%s)" % x.message)
        return rh.reply_fail(rh.REPLY_BAD_REQUEST_400)

    ch_id = _get_new_channel_id()
    if ch_id is None:
        eventlog.log_d('fail: get eventstore ch id')
        return rh.reply_fail(rh.REPLY_INTERNAL_SERVER_ERROR_500)

    # add store_jnode to /etc/edge/setupstore/admin/action/eventstore/<chanel-id>/0.json
    new_jnode_file_path = _get_ch_jnode_path(ch_id)
    try:
        fs.create_jnode_from_default(UI_EVENTSTORE_CHANNEL_DEFJNODE_FILEPATH, new_jnode_file_path)
    except RuntimeError as x:
        eventlog.log_d('fail: create jnode from def (%s)' % x)
        return rh.reply_fail(rh.REPLY_INTERNAL_SERVER_ERROR_500)

    raw_ui_jnode = fs.load_json(new_jnode_file_path)
    if raw_ui_jnode is None:
        return rh.reply_fail(rh.REPLY_INTERNAL_SERVER_ERROR_500)

    # add store to /etc/edge/setupstore/admin/action/eventstore.json
    raw_ui_set_jnode = oh.get_ui_pipeline_eventstore_jnode()
    if raw_ui_set_jnode is None:
        return rh.reply_fail(rh.REPLY_INTERNAL_SERVER_ERROR_500)

    raw_ui_jnode['rw']['args']['name'] = new_jnode['name']
    raw_ui_jnode['ro']['channel'] = ch_id
    bs.append_blobstore(raw_ui_jnode)

    raw_ui_set_jnode.append({
        'internal': False,
        'name': new_jnode['name'],
        'id': ch_id
    })

    with open(new_jnode_file_path, 'w') as json_file:
        ujson.dump(raw_ui_jnode, json_file, indent=2)

    oh.set_ui_pipeline_eventstore_jnode(raw_ui_set_jnode)

    return rh.reply_done(raw_ui_set_jnode)


def _handle_delete_eventstore(query_args):
    # DELETE /api/v1/admin/actions/eventstore/<id>
    if query_args.raw_body:
        eventlog.log_d('fail: bad-request')
        return rh.reply_fail(rh.REPLY_BAD_REQUEST_400)

    _, channel_id = query_args.path_info

    raw_ui_set_jnode = oh.get_ui_pipeline_eventstore_jnode()
    if raw_ui_set_jnode is None:
        return rh.reply_fail(rh.REPLY_INTERNAL_SERVER_ERROR_500)

    store_jnode_set = filter(lambda item_jnode: item_jnode['id'] == channel_id, raw_ui_set_jnode)
    if len(store_jnode_set) != 1:
        eventlog.log_d("fail: internal-server-error")
        return rh.reply_fail(rh.REPLY_INTERNAL_SERVER_ERROR_500)

    if store_jnode_set[0].get('internal', True):
        eventlog.log_d("fail: not allowed to delete internal pipeline")
        return rh.reply_fail(rh.REPLY_METHOD_NOT_ALLOWED_405)

    new_store_set_jnode = filter(lambda pipeline_jnode: pipeline_jnode['id'] != channel_id, raw_ui_set_jnode)

    fs.silent_rm_r(fs.join_path(UI_EVENTSTORE_BASEDIR, channel_id))
    bs.remove_blobstore(channel_id)

    oh.set_ui_pipeline_eventstore_jnode(new_store_set_jnode)

    return rh.reply_done(new_store_set_jnode)


def run(query_args):
    """
    :type query_args: ui_types.UiQueryArgs
    :rtype: dict
    """
    if query_args.http_method == 'GET':
        return _handle_get(query_args)
    
    if query_args.path_info[0] == 'eventstore':
        if query_args.query_string:
            eventlog.log_d('fail: bad-request')
            return rh.reply_fail(rh.REPLY_BAD_REQUEST_400)
        
        if query_args.http_method == 'PUT':
            return _handle_put_eventstore(query_args)
    
        if query_args.http_method == 'DELETE':
            return _handle_delete_eventstore(query_args)
    
        if query_args.http_method == 'POST':
            return _handle_post_eventstore(query_args)

    return rh.reply_fail(rh.REPLY_METHOD_NOT_ALLOWED_405)


def _main():
    eventlog.open_log('cgi-eventstore', 'debug')
    exit_code = 1
    # noinspection PyBroadException
    try:
        # _get_args()
        query_args = ui_types.UiQueryArgs.deserialize()
        reply_jnode = run(query_args)
        print ujson.dumps(reply_jnode)
        exit_code = 0

    except Exception:
        print ujson.dumps(rh.reply_fail(rh.REPLY_INTERNAL_SERVER_ERROR_500))

    finally:
        eventlog.close_log()

    sys.exit(exit_code)


if __name__ == '__main__':
    _main()
