#!/usr/bin/python

import ujson
import subprocess

import ui_pipeline_set
import edge.pipeline_helpers as pipeline_helpers
from edge import path_helpers

_VENDOR_POINTGREY = 'pointgrey'
_VENDOR_BASLER = 'basler'

_MGINFO_POINTGREY_COMMANDLINE = '/opt/edge/bin/mediagrabber-info --vendor=' + _VENDOR_POINTGREY
_MGINFO_BASLER_COMMANDLINE = '/opt/edge/bin/mediagrabber-info --vendor=' + _VENDOR_BASLER

_CAMERA_STATUS_USED = "used"
_CAMERA_STATUS_UNUSED = "unused"
_CAMERA_STATUS_UNDEFINED = "undefined"


def _get_camera_set(mediagrabber_info):
    """
    :type mediagrabber_info: str
    :rtype: list
    """
    cameras = mediagrabber_info.split("===================================")[1:]
    camera_info_set = []

    for camera_info in cameras:
        camera_params = camera_info.split("\n")[1:]
        param_set = map(lambda param: param.strip(), camera_params)
        camera_info_set.append(param_set)

    return camera_info_set


def _check_camera_unused(camera_vendor, camera_serial):
    """
    :type camera_vendor: str
    :type camera_serial: str
    :rtype: bool
    """
    if camera_vendor == _VENDOR_POINTGREY:
        mginfo_commandline = _MGINFO_POINTGREY_COMMANDLINE
    elif camera_vendor == _VENDOR_BASLER:
        mginfo_commandline = _MGINFO_BASLER_COMMANDLINE
    else:
        raise RuntimeError("fail: unknown camera-vendor")

    mg_info = subprocess.Popen([". /opt/edge/rc.setenv; %s --serial=%s" %
                                (mginfo_commandline, camera_serial)],
                               stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)

    mg_stderr = mg_info.stderr.read()
    if mg_stderr:
        raise RuntimeError('fail: mediagrabber-info (%s)' % mg_stderr)

    mg_stdout = mg_info.stdout.read()

    try:
        ujson.loads(mg_stdout)
        return True

    except ValueError:
        pass  # camera status undefined if it is not found in vs-config AND camera info is not a valid JSON

    except:
        raise

    return False


def _check_camera_used(vs_jnode_filepath, camera_serial):
    """
    :type vs_jnode_filepath: str
    :type camera_serial: str
    :rtype: bool
    """
    with open(vs_jnode_filepath, 'r') as json_file:
        videosource_jnode = ujson.load(json_file)
    if videosource_jnode['rw']['type'] == 'mv':
        return camera_serial == videosource_jnode['rw']['serial']


def _get_camera_status(camera_vendor, camera_serial):
    """
    :type camera_vendor: str
    :type camera_serial: str
    :rtype: str
    """
    # get pipeline ids
    pipeline_jnode_set = pipeline_helpers.get_pipeline_jnode_set()
    if pipeline_jnode_set is None:
        raise RuntimeError('fail: get pipeline-jnode-set')

    vs_jnode_filepath_set = \
        map(lambda pid: path_helpers.get_ui_pipeline_jnode_filepath(pid, path_helpers.UI_TYPE_VIDEOSOURCE, '0'),
            map(lambda pipeline: pipeline['id'], pipeline_jnode_set))

    for vs_jnode_filepath in vs_jnode_filepath_set:
        if _check_camera_used(vs_jnode_filepath, camera_serial):
            return _CAMERA_STATUS_USED

    if _check_camera_unused(camera_vendor, camera_serial):
        return _CAMERA_STATUS_UNUSED

    return _CAMERA_STATUS_UNDEFINED


def discover_cameras():
    """
    :rtype: dict
    """
    mg_info = subprocess.Popen([". /opt/edge/rc.setenv; %s" % _MGINFO_POINTGREY_COMMANDLINE],
                               stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)

    mg_stderr = mg_info.stderr.read()
    if mg_stderr:
        raise RuntimeError('fail: mediagrabber-info (%s)' % mg_stderr)
    mg_stdout = mg_info.stdout.read()

    camera_set = _get_camera_set(mg_stdout)

    pg_camera_jnode_set = []
    for camera in camera_set:
        model = camera[2].split(":")[1].strip()
        serial = camera[0].split(":")[1].strip()
        status = _get_camera_status(_VENDOR_POINTGREY, serial)

        pg_camera_jnode = {
            "vendor": _VENDOR_POINTGREY,
            "model": model,
            "serial": serial,
            "status": status
        }

        pg_camera_jnode_set.append(pg_camera_jnode)

    mg_info = subprocess.Popen([". /opt/edge/rc.setenv; %s" % _MGINFO_BASLER_COMMANDLINE],
                               stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)

    mg_stderr = mg_info.stderr.read()
    if mg_stderr:
        raise RuntimeError('fail: mediagrabber-info (%s)' % mg_stderr)

    mg_stdout = mg_info.stdout.read()

    camera_set = _get_camera_set(mg_stdout)

    basler_camera_jnode_set = []
    for camera in camera_set:
        model = camera[2].split(":")[1].strip()
        serial = camera[0].split(":")[1].strip()
        status = _get_camera_status(_VENDOR_BASLER, serial)

        basler_camera_jnode = {
            "vendor": _VENDOR_BASLER,
            "model": model,
            "serial": serial,
            "status": status
        }

        basler_camera_jnode_set.append(basler_camera_jnode)

    camera_jnode_set = []
    if pg_camera_jnode_set:
        camera_jnode_set += pg_camera_jnode_set
    if basler_camera_jnode_set:
        camera_jnode_set += basler_camera_jnode_set

    return camera_jnode_set


