import uuid
import datetime
import ujson
import os

from edge import eventlog
from edge import path_helpers as ph
from edge import pipeline_helpers as oh
from edge.admin import reply_helpers as rh
from edge.admin import ui_types as ui
from edge.com import fs

import ui_pipeline_set
import edge_creator as creator
from _edge_setupstore import append_ss_group
from ui_pipeline import delete_pipeline, get_pipeline_ext_jnode

_TIME_DEF_JNODE_FILEPATH = '/opt/edge/share/setupstore/ui/time-jnode.json'

_UI_IDENTITY_TO_DATASOURCE_MAP = [
    (ph.Pipeline.Mv.Ui.EVENTSTORE_0, False),
    (ph.Pipeline.Mv.Ui.VIDEOSOURCE_0, True),
    (ph.Pipeline.Mv.Ui.LENS_0, True),
    (ph.Pipeline.Mv.Ui.GEOLOCATION_0, False),
    (ph.Pipeline.Mv.Ui.LPR_0, False),
    (ph.Pipeline.Mv.Ui.LANE_ENFORCEMENT_0, False),
    (ph.Pipeline.Mv.Ui.HTTP_STREAMING_0, False),
    (ph.Pipeline.Mv.Ui.RTSP_STREAMING_0, False),
    (ph.Pipeline.Mv.Ui.LOCATION_0, False)
]

_UI_DETECTORS_TO_INTERNAL_MAP = [
    (ph.Pipeline.Mv.Ui.TRANSIT_0, True)
]


def _create_ui_jnode_set(ui_pid, raw_body, origin_uuid, detectors_list):
    """
    :type ui_pid: str
    :type raw_body: str
    :type origin_uuid: str
    :type detectors_list: list
    :rtype: set of UiIdentity
    """
    args = ujson.loads(raw_body)
    serial = args['serial']

    # read def_jnode
    ui_identity_to_def_jnode_map = dict()
    for (ui_identity, _) in _UI_IDENTITY_TO_DATASOURCE_MAP:
        def_ui_jnode_filepath = ph.get_ui_def_jnode_filepath(ui_identity, raw_body)
        def_ui_jnode = fs.load_json2(def_ui_jnode_filepath)
        ui_identity_to_def_jnode_map[ui_identity] = def_ui_jnode

    # read detectors def jnode
    for (ui_identity, internal) in _UI_DETECTORS_TO_INTERNAL_MAP:
        def_ui_jnode_filepath = ph.get_ui_def_jnode_filepath(ui_identity, None)
        def_ui_jnode = fs.load_json2(def_ui_jnode_filepath)
        ui_identity_to_def_jnode_map[ui_identity] = def_ui_jnode
        detector_def_name = def_ui_jnode['rw'].get('args', {}).get('name', '')
        detectors_list.append((ui_identity, internal, detector_def_name))

    # edit
    oh.mk_origin_about(ui_pid, origin_uuid, time=datetime.datetime.now().isoformat(), place=ui_pid)

    ui_identity_to_def_jnode_map[ph.Pipeline.Mv.Ui.EVENTSTORE_0]['rw']['channel'] = int(ui_pid)
    ui_identity_to_def_jnode_map[ph.Pipeline.Mv.Ui.EVENTSTORE_0]['ro']['uuid'] = origin_uuid
    ui_identity_to_def_jnode_map[ph.Pipeline.Mv.Ui.EVENTSTORE_0]['ro']['ctime'] = datetime.datetime.now().isoformat()
    ui_identity_to_def_jnode_map[ph.Pipeline.Mv.Ui.EVENTSTORE_0]['rw']['origin-name'] = ui_pid

    ui_identity_to_def_jnode_map[ph.Pipeline.Mv.Ui.VIDEOSOURCE_0]['rw']['serial'] = str(serial)

    ui_identity_to_def_jnode_map[ph.Pipeline.Mv.Ui.RTSP_STREAMING_0]['rw']['streaming']['tcp-port'] += int(ui_pid)

    eventlog.log_d(ujson.dumps(ui_identity_to_def_jnode_map[ph.Pipeline.Mv.Ui.VIDEOSOURCE_0]))

    # write
    ui_identity_set = set()
    for ui_identity, ui_jnode in ui_identity_to_def_jnode_map.iteritems():
        ui_identity_set.add(ui_identity)
        ui_jnode_dirpath = ph.get_ui_pipeline_jnode_dirpath(ui_pid, ui_identity)
        if not os.path.exists(ui_jnode_dirpath):
            fs.silent_mkdir_r(ui_jnode_dirpath)

        ui_jnode_filepath = ph.get_ui_pipeline_jnode_filepath(ui_pid, ui_identity.type, ui_identity.id)
        with open(ui_jnode_filepath, 'w') as json_file:
            ujson.dump(ui_jnode, json_file, indent=2)

        ui_schema_filepath = ph.get_ui_pipeline_schema_filepath(ui_pid, ui_identity.type, ui_identity.id)
        fs.ln_schema(ui_schema_filepath, ui_jnode_filepath)

    return ui_identity_set


def create_mv_pipeline(body, ui_pid):
    """
    :type body: str
    :type ui_pid: str
    """
    entry_point = ph.Pipeline.Mv.Ui.PIPELINE
    pipeline_jnode_set = oh.get_pipeline_jnode_set2()
    if pipeline_jnode_set is None:
        eventlog.log_d("fail: internal-server-error")
        return rh.reply_fail(rh.REPLY_INTERNAL_SERVER_ERROR_500)

    # noinspection PyBroadException
    try:
        origin_uuid = str(uuid.uuid4())
        origin_name = ui_pid
        # create ui-jnode-set
        detectors_list = []
        ui_identity_set = _create_ui_jnode_set(ui_pid, body, origin_uuid, detectors_list)
        # create edge-jnode-set
        creator.generate(ui_pid, entry_point, None, ui_identity_set)

        # update pipeline.json
        visible_ui_type_set = map(lambda (module, datasource): (module.type, datasource),
                                  _UI_IDENTITY_TO_DATASOURCE_MAP)

        pipeline_jnode = ui_pipeline_set.add_pipeline(
            pipeline_jnode_set, ui_pid, origin_name, origin_uuid,
            entry_point.ptype, visible_ui_type_set,
            ui_detectors_list=detectors_list, internal=False)

        pipeline_ext_jnode = get_pipeline_ext_jnode(pipeline_jnode)

        # create time-jnode
        fs.create_jnode_from_default(_TIME_DEF_JNODE_FILEPATH, ph.UI_TIME_JNODE_FILEPATH)

        # update setupstore config
        append_ss_group(ui_pid, entry_point)
        # restart webserver with PCD
        oh.restart_webserver_lazy()
        oh.sighup_setupstore()

        return rh.reply_done(pipeline_ext_jnode)

    except NotImplementedError:
        eventlog.log_d('fail: camera not supported')
        return rh.reply_fail(rh.REPLY_BAD_REQUEST_400)

    except ui.LockError:
        return rh.reply_fail(rh.REPLY_CONFLICT_409)

    except Exception:
        eventlog.log_x()

    delete_pipeline(ui_pid)
    return rh.reply_fail(rh.REPLY_INTERNAL_SERVER_ERROR_500)
