from edge.admin.reply_helpers import *
from edge.admin.ui_types import UI_PIPELINE_PID_INDEX, UI_PIPELINE_TYPE_INDEX, UI_PIPELINE_PREFIX_INDEX
from edge.path_helpers import *
import edge.pipeline_helpers as ph

import ujson
from os.path import join as join_path

import _camera_discover
import ui_pipeline_set
from _edge_eventstore import EsDeleter
from _edge_grabcontrol import GcDeleter
from _edge_lpr import LrDeleter
from _edge_rawcodec import RcDeleter
from _edge_mediagrabber import MgDeleter
from _edge_radar import RaDeleter
from _edge_renderer import OrDeleter
from _edge_rtspgrabber import RgDeleter
from _edge_rtspserver import RsDeleter
from _edge_setupstore import remove_ss_group
from _edge_speedcontrol import ScDeleter
from _edge_tracker import TrDeleter
from _edge_webserver import WsDeleter
from _edge_mediacodecgst import McGstDeleter
from _edge_blobstore import BsDeleter
from _edge_imagemapper import ImDeleter
from _edge_trafficlight import TlDeleter

from _ui_eventstore import UiEventStore
from _ui_geolocation import UiGeoLocation
from _ui_http_streaming import UiHttpStreaming
from _ui_lane_enforcement import UiLaneEnforcement
from _ui_lens import UiLens
from _ui_location import UiLocation
from _ui_lpr import UiLpr
from _ui_rtsp_streaming import UiRtspStreaming
from _ui_videosource import UiVideosource
from _ui_transit import UiTransit
from _ui_parking import UiParking
from _ui_spot_speed import UiSpotSpeed
from _ui_wrong_way import UiWrongWay
from _ui_red_light import UiRedLight
import edge.pipeline_helpers as helpers
from ui_pipeline_set import get_pipeline_type

_UI_TYPE_TO_HANDLER_CTOR_MAP = {
    UI_TYPE_LOCATION: UiLocation.create,
    UI_TYPE_GEOLOCATION: UiGeoLocation.create,
    UI_TYPE_VIDEOSOURCE: UiVideosource.create,
    UI_TYPE_LPR: UiLpr.create,
    UI_TYPE_EVENTSTORE: UiEventStore.create,
    UI_TYPE_TRANSIT: UiTransit.create,
    UI_TYPE_RTSP_STREAMING: UiRtspStreaming.create,
    UI_TYPE_HTTP_STREAMING: UiHttpStreaming.create,
    UI_TYPE_LENS: UiLens.create,
    UI_TYPE_LANE_ENFORCEMENT: UiLaneEnforcement.create,
    UI_TYPE_PARKING: UiParking.create,
    UI_TYPE_SPOT_SPEED: UiSpotSpeed.create,
    UI_TYPE_WRONG_WAY: UiWrongWay.create,
    UI_TYPE_RED_LIGHT: UiRedLight.create
}

_EGDE_TYPE_TO_DELETER_CTOR_MAP = [
    MgDeleter.create,
    GcDeleter.create,
    RgDeleter.create,
    OrDeleter.create,
    RsDeleter.create,

    RaDeleter.create,
    LrDeleter.create,
    TrDeleter.create,
    ScDeleter.create,
    EsDeleter.create,
    WsDeleter.create,
    RcDeleter.create,
    McGstDeleter.create,
    BsDeleter.create,
    ImDeleter.create,
    TlDeleter.create
]


def delete_pipeline(ui_pid):
    """
    :type ui_pid: str
    :rtype: dict
    """
    ptype = get_pipeline_type(ui_pid)
    if not ptype:
        # pipeline is not defined
        return

    if ptype == PTYPE_LPR_RTSP:
        entry_point = Pipeline.Rtsp.Ui.PIPELINE
    elif ptype == PTYPE_LPR_MV:
        entry_point = Pipeline.Mv.Ui.PIPELINE
    elif ptype == PTYPE_SPEED:
        entry_point = Pipeline.Speed.Ui.PIPELINE
    else:
        raise NotImplemented

    origin_name = ph.get_origin_name(ui_pid) or ''

    # rm configs of pipeline-edge-modules and detach videostreams of webserver
    deleter_set = filter(bool, map(lambda ctor: ctor(ui_pid, entry_point),
                                   _EGDE_TYPE_TO_DELETER_CTOR_MAP))

    map(lambda deleter: deleter.run(), deleter_set)

    # rm configs of ui-modules
    shemas_dirpath = join_path(BASE_UI_PIPELINE_SCHEMA_DIRPATH, ui_pid)
    ui_dirpath = join_path(BASE_UI_PIPELINE_JNODE_DIRPATH, ui_pid)

    map(helpers.silent_rm_r, [shemas_dirpath, ui_dirpath])
    ui_pipelines_jnode = ui_pipeline_set.remove_pipeline(ui_pid, origin_name)
    remove_ss_group(ui_pid)
    return ui_pipelines_jnode


def get_pipeline_ext_jnode(pipeline_jnode_list):
    """
    :type pipeline_jnode_list: list of dict
    :rtype: dict
    """
    for pipeline_jnode in pipeline_jnode_list:
        ui_pid = pipeline_jnode['id']
        pipeline_modules_jnode = pipeline_jnode['modules']
        videosource_module_jnode = filter(lambda module: module['type'] == UI_TYPE_VIDEOSOURCE, pipeline_modules_jnode)
        # noinspection PyTypeChecker
        ui_id = videosource_module_jnode[0]['id']

        videosource_error = False
        videsource_status = helpers.get_videsource_status(ui_pid, ui_id)
        if videsource_status is None:
            eventlog.log_d('fail: get videosource status')
            videosource_error = True

        elif videsource_status != 0:
            eventlog.log_d('fail: videosource status error')
            videosource_error = True

        # set pipeline status
        videosource_dependent_types = [UI_TYPE_VIDEOSOURCE, UI_TYPE_LPR, UI_TYPE_LANE_ENFORCEMENT, UI_TYPE_HTTP_STREAMING]
        for module_jnode in pipeline_modules_jnode:
            if module_jnode['type'] in videosource_dependent_types:
                module_jnode['error'] = videosource_error

            else:
                module_jnode['error'] = False

        detectors_types = ['parking', 'wrong-way', 'red-light']

        # pop ptype from jnode
        ptype = pipeline_jnode.pop('type', None)
        if ptype == PTYPE_SPEED:
            detectors_types.append('spot-speed')

        pipeline_jnode['limits'] = {
            'detectors': detectors_types
        }

    return pipeline_jnode_list


class UiPipeline:
    def __init__(self, handler, handler_args):
        """
        :type handler: object -> dict
        :type handler_args: object
        """
        self._handler = handler
        self._handler_args = handler_args

    @staticmethod
    def _handle_get_pipeline(pipeline_jnode):
        """
        :type pipeline_jnode: dict
        :rtype: dict
        """
        return reply_done(get_pipeline_ext_jnode([pipeline_jnode])[0])

    @staticmethod
    def _handle_define_videosource(_):
        return reply_done(_camera_discover.discover_cameras())

    @staticmethod
    def _handle_delete_pipeline(ui_pid):
        """
        :type ui_pid: str
        :rtype: dict
        """
        ui_pipeline_jnode = delete_pipeline(ui_pid)

        # restarting edgeserver with PCD
        eventlog.log_d('restarting modules')
        helpers.sighup_setupstore()
        helpers.restart_webserver_lazy()

        return reply_done(ui_pipeline_jnode)

    @staticmethod
    def _deserialize_get_pipeline(query_args):
        """
        :type query_args: edge.admin.ui_types.UiQueryArgs
        :rtype: UiPipeline
        """
        if len(query_args.path_info) != 1:
            eventlog.log_d("fail: not-found (%s)" % query_args.path_info)
            return UiPipeline(reply_fail, handler_args=REPLY_NOT_FOUND_404)

        if query_args.path_info:
            if query_args.path_info[0] == 'define-videosource':
                return UiPipeline(UiPipeline._handle_define_videosource, None)

        if query_args.query_string or query_args.raw_body:
            eventlog.log_d("fail: bad-request")
            return UiPipeline(reply_fail, handler_args=REPLY_BAD_REQUEST_400)

        pipeline_set_jnode = helpers.get_pipeline_jnode_set2()
        if not pipeline_set_jnode:
            eventlog.log_d("fail: internal-server-error (invalid pipeline-set-jnode)")
            return UiPipeline(reply_fail, handler_args=REPLY_INTERNAL_SERVER_ERROR_500)

        ui_pid = query_args.get_path_info_item(UI_PIPELINE_PID_INDEX)
        pipeline_jnode_set = filter(lambda item_jnode: item_jnode['id'] == ui_pid, pipeline_set_jnode)

        if not pipeline_jnode_set:
            eventlog.log_d("fail: not-found (%s)" % query_args.path_info[0])
            return UiPipeline(reply_fail, handler_args=REPLY_NOT_FOUND_404)

        elif len(pipeline_jnode_set) > 1:
            eventlog.log_d("fail: internal-server-error")
            return UiPipeline(reply_fail, handler_args=REPLY_INTERNAL_SERVER_ERROR_500)

        return UiPipeline(UiPipeline._handle_get_pipeline, pipeline_jnode_set[0])

    @staticmethod
    def _deserialize_delete_pipeline(query_args):
        """
        :type query_args: edge.admin.ui_types.UiQueryArgs
        :rtype: UiPipeline
        """
        if query_args.query_string or query_args.raw_body:
            eventlog.log_d("fail: bad-request")
            return UiPipeline(reply_fail, handler_args=REPLY_BAD_REQUEST_400)

        pipeline_set_jnode = helpers.get_pipeline_jnode_set()
        if not pipeline_set_jnode:
            eventlog.log_d("fail: internal-server-error (invalid pipeline-set-jnode)")
            return UiPipeline(reply_fail, handler_args=REPLY_INTERNAL_SERVER_ERROR_500)

        ui_pid = query_args.get_path_info_item(UI_PIPELINE_PID_INDEX)
        pipeline_jnode_set = filter(lambda item_jnode: item_jnode['id'] == ui_pid, pipeline_set_jnode)
        if len(pipeline_jnode_set) != 1:
            eventlog.log_d("fail: internal-server-error")
            return UiPipeline(reply_fail, handler_args=REPLY_INTERNAL_SERVER_ERROR_500)

        if pipeline_jnode_set[0].get('internal', True):
            eventlog.log_d("fail: not allowed to delete internal pipeline")
            return UiPipeline(reply_fail, handler_args=REPLY_METHOD_NOT_ALLOWED_405)

        return UiPipeline(UiPipeline._handle_delete_pipeline, ui_pid)

    @staticmethod
    def _deserialize_post_pipeline(query_args):
        """
        :type query_args: edge.admin.ui_types.UiQueryArgs
        :rtype: UiPipeline
        """
        ui_prefix = query_args.get_path_info_item(UI_PIPELINE_PREFIX_INDEX)
        if ui_prefix != 'detectors':
            eventlog.log_d("fail: not-found")
            return UiPipeline(reply_fail, handler_args=REPLY_NOT_FOUND_404)

        raw_body = ujson.loads(query_args.raw_body)
        ui_type = raw_body['type']
        ui_handler_ctor = _UI_TYPE_TO_HANDLER_CTOR_MAP.get(ui_type)
        if not ui_handler_ctor:
            eventlog.log_d("fail: not-found")
            return UiPipeline(reply_fail, handler_args=REPLY_NOT_FOUND_404)

        ui_handler = ui_handler_ctor(query_args)

        return UiPipeline(handler=lambda args: args.handle(), handler_args=ui_handler)

    @staticmethod
    def deserialize(query_args):
        """
        :type query_args: edge.admin.ui_types.UiQueryArgs
        :rtype: UiPipeline
        """
        if not query_args.get_ui_type():
            if query_args.http_method == 'GET':
                return UiPipeline._deserialize_get_pipeline(query_args)

            if query_args.http_method == 'DELETE':
                return UiPipeline._deserialize_delete_pipeline(query_args)

            if query_args.http_method == 'POST':
                return UiPipeline._deserialize_post_pipeline(query_args)

            eventlog.log_d("method-not-allowed")
            return UiPipeline(reply_fail, handler_args=REPLY_METHOD_NOT_ALLOWED_405)

        ui_prefix = query_args.get_path_info_item(UI_PIPELINE_PREFIX_INDEX)
        if ui_prefix not in ['detectors', 'devices']:
            eventlog.log_d("fail: not-found")
            return UiPipeline(reply_fail, handler_args=REPLY_NOT_FOUND_404)

        ui_type = query_args.get_path_info_item(UI_PIPELINE_TYPE_INDEX)
        ui_handler_ctor = _UI_TYPE_TO_HANDLER_CTOR_MAP.get(ui_type)
        if not ui_handler_ctor:
            eventlog.log_d("fail: not-found")
            return UiPipeline(reply_fail, handler_args=REPLY_NOT_FOUND_404)

        ui_handler = ui_handler_ctor(query_args)

        return UiPipeline(handler=lambda args: args.handle(), handler_args=ui_handler)

    def handle(self):
        """
        :rtype: dict
        """
        return self._handler(self._handler_args)
