#!/usr/bin/python
import os

import iso8601
import jsonschema
import datetime
import ujson

from edge import eventlog
from edge.admin.cgi import run_cgi, CgiModule, CgiLock, CgiBackup
from edge.admin.cgi import CgiInternalServerError, CgiBadRequestError, CgiNotFoundError, CgiPathInfo
from edge.admin.reply_helpers import Reply

from edge.com import sh, net, fs
from edge.pipeline_helpers import restart_webserver_lazy


class _Assets:
    UI_SCHEMA = '/etc/edge/setupstore/schemas/time.json'
    UI_SCHEMA_MANUAL = '/opt/edge/share/setupstore/ui/time-schema-manual.json'
    UI_JNODE = '/etc/edge/setupstore/admin/actions/time.json'
    UI_DEF_JNODE = '/opt/edge/share/setupstore/ui/time-jnode.json'

    NTP_CONF = '/etc/ntp.conf'
    NTPDATE_CONF = '/etc/default/ntpdate'

    GNSS_LIMIT_JNODE = {
        "type": "gnss",
        "sync-auto": True,
        "sync-manual": False
    }

    def __init__(self):
        raise NotImplementedError


class _Module(CgiModule):
    """
    GET     /api/v2/admin/actions/time
    PUT     /api/v2/admin/actions/time
    PUT     /api/v2/admin/actions/time  /manual
    """
    def __init__(self, handler):
        """
        :type handler: () -> Reply
        """
        CgiModule.__init__(self, handler)

    @classmethod
    def unpack(cls, request):
        """
        :type request: CgiRequest
        :rtype: CgiModule

        :raise CgiNotFoundError:
        :raise CgiBadRequestError:
        """
        method_to_handlers_map = {
            'GET': [
                cls._unpack_get_time
            ],
            'PUT': [
                cls._unpack_put_time,
                cls._unpack_put_time_manual
            ]
        }

        handlers = method_to_handlers_map.get(request.method, [])
        for handler in handlers:
            try:
                return handler(request)

            except CgiBadRequestError:
                raise

            except CgiNotFoundError:
                pass

        eventlog.log_i('fail: http404')
        raise CgiNotFoundError

    # noinspection PyUnusedLocal
    @staticmethod
    def _unpack_get_time(request, limits=None):
        """
        :type request: CgiRequest
        :type limits: dict | None
        :rtype: CgiModule

        :raise CgiNotFoundError:
        :raise CgiBadRequestError:
        """
        #
        path_constraints = [
            len(request.path_info) == 0,
        ]

        if not all(path_constraints):
            raise CgiNotFoundError

        #
        args_constraints = [
            not request.query_string,
            not request.body
        ]

        if not all(args_constraints):
            eventlog.log_i('fail: http400')
            raise CgiBadRequestError

        return CgiModule(_handle_get_time)

    # noinspection PyUnusedLocal
    @staticmethod
    def _unpack_put_time(request, limits=None):
        """
        :type request: CgiRequest
        :type limits: dict | None
        :rtype: CgiModule

        :raise CgiNotFoundError:
        :raise CgiBadRequestError:
        """
        #
        path_constraints = [
            len(request.path_info) == 0,
        ]

        if not all(path_constraints):
            raise CgiNotFoundError

        #
        args_constraints = [
            not request.query_string,
            request.body
        ]

        if not all(args_constraints):
            eventlog.log_i('fail: http400')
            raise CgiBadRequestError

        # define next-jnode
        try:
            next_jnode = ujson.loads(request.body)

        except ValueError as x:
            eventlog.log_i("fail: http400. loads jnode <%s>" % x.message)
            raise CgiBadRequestError

        # define next-schema
        try:
            with open(_Assets.UI_SCHEMA, 'r') as json_file:
                next_schema = ujson.load(json_file)

        except (IOError, ValueError) as x:
            eventlog.log_e("fail: http500. load schema <%s>" % x.message)
            raise CgiInternalServerError

        # validate next-jnode against next-schema
        try:
            jsonschema.validate(next_jnode, next_schema)

        except (ValueError, jsonschema.ValidationError) as x:
            eventlog.log_i("fail: http400. validate jnode <%s>" % x.message)
            raise CgiBadRequestError

        return CgiModule(lambda: _handle_put_time(next_jnode['args']))

    # noinspection PyUnusedLocal
    @staticmethod
    def _unpack_put_time_manual(request, limits=None):
        """
        :type request: CgiRequest
        :type limits: dict | None
        :rtype: CgiModule

        :raise CgiNotFoundError:
        :raise CgiBadRequestError:
        """
        #
        path_constraints = [
            len(request.path_info) == 1,
            request.path_info[0] == 'manual',
        ]

        if not all(path_constraints):
            raise CgiNotFoundError

        #
        args_constraints = [
            not request.query_string,
            request.body
        ]

        if not all(args_constraints):
            eventlog.log_i('fail: http400')
            raise CgiBadRequestError

        # define jnode
        try:
            jnode = ujson.loads(request.body)

        except ValueError as x:
            eventlog.log_i("fail: http400 <%s>" % x.message)
            raise CgiBadRequestError

        # define schema
        try:
            with open(_Assets.UI_SCHEMA_MANUAL, 'r') as json_file:
                schema = ujson.load(json_file)

        except (IOError, ValueError) as x:
            eventlog.log_e("fail: http500 <%s>" % x.message)
            raise CgiInternalServerError

        # validate jnode against schema
        try:
            jsonschema.validate(jnode, schema)

        except (ValueError, jsonschema.ValidationError) as x:
            eventlog.log_i("fail: http400 <%s>" % x.message)
            raise CgiBadRequestError

        return CgiModule(lambda: _handle_put_time_manual(sync_source=jnode))


def _edit_ntp_conf(conf_path, ui_jnode, next_ntp_source, next_gnss_source):
    """
    :type ui_jnode: dict
    :type next_ntp_source: dict
    :type next_gnss_source: dict
    :rtype: bool

    :raise CgiInternalServerError:
    """
    ui_ro = ui_jnode['ro']
    conf_lines = []

    autosync = False
    if next_ntp_source.get('autosync', False):
        autosync = True
        for server_jnode in next_ntp_source['servers']:
            address = server_jnode['address']
            if server_jnode['prefered']:
                conf_lines.append('server %s prefer' % address)
            else:
                conf_lines.append('server %s' % address)

        for server_jnode in next_ntp_source['servers']:
            conf_lines.append('restrict %s noquery notrap' % server_jnode['address'])

    if next_gnss_source.get('autosync', False):
        autosync = True
        gnss_conf_lines = ui_ro.get('ntp-conf-gnss-vit', [])
        if not gnss_conf_lines:
            gnss_conf_lines = ui_ro.get('ntp-conf-gnss')
            if gnss_conf_lines is None:
                eventlog.log_e_param('ui#/ro/ntp-conf-gnss')
                raise CgiInternalServerError

        conf_lines.extend(gnss_conf_lines)

    conf_lines.extend(ui_ro.get('ntp-conf-footer', []))

    with open(conf_path, 'w') as conf_file:
        conf_file.write('\n'.join(conf_lines))

    return autosync


def _edit_ntpdate_conf(conf_path, ntp_autosync):
    """
    :type ntp_autosync: bool
    """
    with open(conf_path, 'r') as conf_file:
        last_lines = conf_file.readlines()

    next_lines = filter(lambda line: not line.startswith('exit 0'), last_lines)
    if ntp_autosync:
        # do sync time using ntpdate on system startup
        pass
    else:
        # do not sync time using ntpdate on system startup
        next_lines.insert(0, 'exit 0\n')

    with open(conf_path, 'w') as conf_file:
        conf_file.write(''.join(next_lines))


def _sync_user_manual(sync_source):
    """
    :type sync_source: dict
    :raise CgiInternalServerError:
    """
    tz_datetime = iso8601.parse_date(sync_source['time'])
    utc_datetime = tz_datetime - tz_datetime.utcoffset()

    # set OS time
    shell_cmd = 'date --set="{:%Y-%m-%d %H:%M:%S}"'.format(utc_datetime)
    answer = sh.run_silent(shell_cmd)
    if answer == 0:
        eventlog.log_i('done: sh=<{}>'.format(shell_cmd))
    else:
        eventlog.log_e('fail: http500. sh=<{}> rc=<{}>'.format(shell_cmd, answer))
        raise CgiInternalServerError

    # sync BIOS time to OS time
    shell_cmd = 'hwclock -w'
    answer = sh.run_silent(shell_cmd)
    if answer == 0:
        eventlog.log_i('done: sh=<{}>'.format(shell_cmd))
    else:
        eventlog.log_e('fail: http500. sh=<{}> rc=<{}>'.format(shell_cmd, answer))
        raise CgiInternalServerError


def _sync_ntp_manual(sync_source):
    """
    :type sync_source: dict
    :raise CgiInternalServerError:
    """
    # set OS time
    servers = map(lambda server: sh.quote(server['address']), sync_source.get('servers', []))
    if not servers:
        eventlog.log_e('fail: http500')
        raise CgiInternalServerError

    shell_cmd = 'ntpdate -u {}'.format(' '.join(servers))
    answer = sh.run_silent(shell_cmd)
    if answer == 0:
        eventlog.log_i('done: sh=<{}>'.format(shell_cmd))
    else:
        eventlog.log_e('fail: sh=<{}> rc=<{}>'.format(shell_cmd, answer))
        raise CgiInternalServerError

    # sync BIOS time to OS time
    shell_cmd = 'hwclock -w'
    answer = sh.run_silent(shell_cmd)
    if answer == 0:
        eventlog.log_i('done: sh=<{}>'.format(shell_cmd))
    else:
        eventlog.log_e('fail: sh=<{}> rc=<{}>'.format(shell_cmd, answer))
        raise CgiInternalServerError


def _handle_get_time():
    """
    :rtype: Reply
    :raise CgiInternalServerError:
    """

    # create ui-jnode
    if not os.path.exists(_Assets.UI_JNODE):
        # noinspection PyBroadException
        try:
            fs.create_jnode_from_default(_Assets.UI_DEF_JNODE, _Assets.UI_JNODE)
        except Exception:
            eventlog.log_e('fail: http500. mk file <{}>'.format(_Assets.UI_JNODE))
            eventlog.log_x()
            raise CgiInternalServerError

        if not os.path.exists(_Assets.UI_JNODE):
            eventlog.log_e('fail: http500. file absent <{}>'.format(_Assets.UI_JNODE))
            raise CgiInternalServerError

    # create ui-jnode
    if not os.path.islink(_Assets.UI_SCHEMA) or not os.path.exists(os.readlink(_Assets.UI_SCHEMA)):
        fs.ln_schema(_Assets.UI_SCHEMA, _Assets.UI_JNODE)

    try:
        ui_jnode = fs.load_json2(_Assets.UI_JNODE)
        ui_rw = ui_jnode['rw']
        if net.query_gps_count() > 0:
            ui_rw['limits']['sources'].append(_Assets.GNSS_LIMIT_JNODE)

        else:
            ui_rw['args']['sources'] = filter(
                lambda source: source['type'] != 'gnss',
                ui_rw['args']['sources'])

        return Reply(body=ui_rw)
    except fs.FsError:
        eventlog.log_e('fail: http500')
        raise CgiInternalServerError


@CgiLock('/var/run/edge/datetime')
@CgiBackup('/etc/edge/setupstore/backup/datetime', {
    'time.json': _Assets.UI_JNODE,
    'ntp.conf': _Assets.NTP_CONF,
    'ntpdate': _Assets.NTPDATE_CONF
})
def _handle_put_time(body_args):
    """
    :type body_args: dict
    :rtype: Reply

    :raise CgiInternalServerError:
    :raise CgiBadRequestError:
    """
    try:
        with open(_Assets.UI_JNODE, 'r') as json_file:
            ui_jnode = ujson.load(json_file)

    except (IOError, ValueError) as x:
        eventlog.log_e("fail: http500. load <{}> <{}>".format(_Assets.UI_JNODE, x.message))
        raise CgiInternalServerError

    ui_jnode_rw = ui_jnode['rw']
    if body_args['tz'] not in ui_jnode_rw['limits']['tz']:
        eventlog.log_i("fail: http400. param #/args/tz has invalid value")
        raise CgiBadRequestError

    def get_gnss_sync_source(sync_sources):
        """
        :param sync_sources:
        :rtype: dict

        :raise CgiBadRequestError:
        """
        gnss_sources = filter(lambda source: source['type'] == 'gnss', sync_sources)
        gnss_sources_count = len(gnss_sources)
        if gnss_sources_count == 1:
            return gnss_sources[0]

        elif gnss_sources_count == 0:
            return {
                'autosync': False,
                'type': 'gnss'
            }

        else:
            eventlog.log_i("fail: http400. param #/args/sources has invalid value")
            raise CgiBadRequestError

    def get_ntp_sync_source(sync_sources):
        """
        :type sync_sources: list[dict]
        :rtype: dict

        :raise CgiBadRequestError:
        """
        ntp_sources = filter(lambda source: source['type'] == 'ntp', sync_sources)
        ntp_sources_count = len(ntp_sources)
        if ntp_sources_count == 1:
            return ntp_sources[0]

        elif ntp_sources_count == 0:
            return {
                'autosync': False,
                'type': 'ntp',
                'servers': []
            }

        else:
            eventlog.log_i("fail: http400. param #/args/sources has invalid value")
            raise CgiBadRequestError

    next_gnss_source = get_gnss_sync_source(body_args.get('sources', []))
    next_ntp_source = get_ntp_sync_source(body_args.get('sources', []))

    _edit_ntpdate_conf(_Assets.NTPDATE_CONF, next_ntp_source.get('autosync', False))
    autosync = _edit_ntp_conf(_Assets.NTP_CONF, ui_jnode, next_ntp_source, next_gnss_source)

    # restart ntpd
    sh.run_silent('service ntp stop')
    sh.run_silent('update-rc.d ntp disable')
    if autosync:
        sh.run_silent('update-rc.d ntp enable')
        sh.run_silent('service ntp start')

    # if tz changed restart
    if ui_jnode_rw['args']['tz'] != body_args['tz']:
        eventlog.log_i('exec: restart overlayrender, speedcontrol, webserver')
        sh.run_silent('pkill -9 overlayrender')
        sh.run_silent('pkill -9 speedcontrol')
        restart_webserver_lazy()
        # TODO: restart edgeserver if next_time < current_time

    # write new jnode to disk
    ui_jnode_rw['args'] = body_args
    with open(_Assets.UI_JNODE, 'w') as json_file:
        ujson.dump(ui_jnode, json_file, indent=2)

    if net.query_gps_count() > 0:
        ui_jnode_rw['limits']['sources'].append(_Assets.GNSS_LIMIT_JNODE)

    return Reply(body=ui_jnode_rw)


def _handle_put_time_manual(sync_source):
    """
    :type sync_source: dict
    :rtype: Reply
    """
    # handle
    source_type = sync_source.get('type')
    if source_type == 'ntp':
        _sync_ntp_manual(sync_source)

    elif source_type == 'user':
        _sync_user_manual(sync_source)

    else:
        eventlog.log_i('fail: http400. param #/type has invalid value <{}>'.format(source_type))
        raise CgiBadRequestError

    return Reply(body={
        'time': datetime.datetime.now().isoformat()
    })


def _main():
    run_cgi(
        lambda arg: CgiPathInfo.unpack(arg),
        lambda arg: _Module.unpack(arg))


if __name__ == '__main__':
    eventlog.open_log('cgi-time', 'debug')
    try:
        _main()
    finally:
        eventlog.close_log()
