
from edge import eventlog
from edge.image.jnode.types import JsonImEngineArgs
from edge.types import mk_error_text, JnodeError

from ...com.jnode.types import JsonBaseRuleArgs, JsonVideoArgs
from ..types import DetectorArgs, RuleArgs, RuleLimits

from ...types import IDetector
from ....model.types import Identity
from ....model.jnode.types import JsonSceneArgs


class JsonRuleLimits(RuleLimits):
    def __init__(self, *args, **kwargs):
        RuleLimits.__init__(self, *args, **kwargs)

    @staticmethod
    def deserialize(limits_jnode):
        """
        :type limits_jnode: dict
        :rtype: JsonRuleLimits
        :raise JnodeError:
        """
        # deserialize #/max-standing-time
        min_red_time = limits_jnode.get('min-red-time')
        if not (int == type(min_red_time) and min_red_time > 0):
            eventlog.log_d(mk_error_text('#/min-red-time'))
            raise JnodeError

        return JsonRuleLimits(min_red_time)


class JsonRuleArgs(RuleArgs):
    def __init__(self, *args, **kwargs):
        RuleArgs.__init__(self, *args, **kwargs)

    @staticmethod
    def deserialize(rule_jnode):
        """
        :type rule_jnode: dict
        :rtype: JsonRuleArgs | None
        :raise JnodeError:
        """
        if type(rule_jnode) != dict:
            eventlog.log_d(mk_error_text('#/'))
            raise JnodeError

        # deserialize #/name
        # deserialize #/enabled
        # deserialize #/ticket
        rule = JsonBaseRuleArgs.deserialize(rule_jnode)
        if not rule.enabled:
            return None

        # deserialize #/limits
        limits_jnode = rule_jnode.get('limits')
        if type(rule_jnode) != dict:
            eventlog.log_d(mk_error_text('#/limits'))
            raise JnodeError

        limits = JsonRuleLimits.deserialize(limits_jnode)
        return JsonRuleArgs(rule.name, rule.ticket, limits)


class JsonDetectorArgs(DetectorArgs):
    def __init__(self, *args, **kwargs):
        DetectorArgs.__init__(self, *args, **kwargs)

    @staticmethod
    def deserialize(jnode, index, def_footer_mapping):
        """
        :type jnode: dict
        :type index: int
        :type def_footer_mapping: dict
        :rtype: JsonDetectorArgs
        :raise JnodeError:
        """
        # deserialize #/name
        name = jnode.get('name')
        if not type(name) in [str, unicode] or len(name) > 32:
            eventlog.log_d(mk_error_text('#/name'))
            raise JnodeError

        # deserialize #/store
        store = jnode.get('store')
        if not (type(store) in [str, unicode] and 1 <= len(store) <= 2):
            eventlog.log_d(mk_error_text('#/store'))
            raise JnodeError

        # deserialize #/scene
        scene_jnode = jnode.get('scene')
        if type(scene_jnode) != dict:
            eventlog.log_d(mk_error_text('#/scene'))
            raise JnodeError
        scene = JsonSceneArgs.deserialize(scene_jnode)

        # deserialize #/rules
        rules_jnode = jnode.get('rules')
        if type(rules_jnode) != list:
            eventlog.log_d(mk_error_text('#/rules'))
            raise JnodeError

        rules = filter(bool, map(JsonRuleArgs.deserialize, rules_jnode))
        identity = Identity(IDetector.TYPE_RED_LIGHT, index)

        # deserialize #/images
        images_jnode = jnode.get('images')
        if type(images_jnode) != dict:
            eventlog.log_d(mk_error_text('#/images'))
            raise JnodeError

        im_engine = JsonImEngineArgs.deserialize(images_jnode['violation'])

        # deserialize #/videos/measuments
        video_jnode = jnode.get('videos', {}).get('measurement')
        video_args = JsonVideoArgs.deserialize(video_jnode) if video_jnode else None
        return JsonDetectorArgs(identity, name, store, scene, rules, im_engine, def_footer_mapping, video_args)
